using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

using CUZK.MergeDBWithPoints.Addresses;
using CUZK.MergeDBWithPoints.Mapping;
using CUZK.MergeDBWithPoints.Territories;

using CUZK.Common;


namespace CUZK.MergeDBWithPoints {
	class AddressesFinder {
		AddressesDB addresses;
		TerritoriesDB territories;
		Mappings mappings;
		BuildingDefinitionPointDB points;

		System.Globalization.CultureInfo czechCulture;

		public List<AddressPoint> matchedAddressPoints;
		public List<AddressPoint> unmatchedAdressPoints;
		public Dictionary<District, List<BuildingDefinitionPoint>> unmatchedBuildingDefinitionPoints;

		public AddressesFinder(AddressesDB a, TerritoriesDB t, Mappings m, BuildingDefinitionPointDB p) {
			addresses = a;
			territories = t;
			mappings = m;
			points = p;

			czechCulture = System.Globalization.CultureInfo.GetCultureInfo("cs");

			matchedAddressPoints = new List<AddressPoint>();
			unmatchedAdressPoints = new List<AddressPoint>();
			unmatchedBuildingDefinitionPoints = new Dictionary<District, List<BuildingDefinitionPoint>>();
		}


		public void FindAddressesAssignment() {
			Dictionary<string, List<BuildingDefinitionPoint>> unassignedBuildngDefinitionPoints = new Dictionary<string, List<BuildingDefinitionPoint>>();
			Dictionary<District, List<AddressPoint>> unassignedAddressPoints = new Dictionary<District, List<AddressPoint>>();

			// loads lists of BuildingDefinitionPoints a AddressPoints
			Initialize(unassignedBuildngDefinitionPoints, unassignedAddressPoints);

			foreach (var mapping in mappings.MappingsList) {
				try{
				List<BuildingDefinitionPoint> territoryBuildings = unassignedBuildngDefinitionPoints[mapping.TerritoryName];
				List<AddressPoint> terrritoryAddresses = new List<AddressPoint>();

				foreach (var mappingDistrict in mapping.Districts) {
					terrritoryAddresses.AddRange(unassignedAddressPoints[mappingDistrict]);
				}

					foreach (var definitionPoint in territoryBuildings.ToList()) {
						AddressPoint buildingAP = null;

						Regex numberRegex = new Regex("([0-9]+)");
						Match regexResult = numberRegex.Match(definitionPoint.Description);
						if (regexResult.Success == false)
							continue;

						if (IsBuildingWithConscriptionNumber(definitionPoint)) {
							string conscriptionNumber = regexResult.Groups[0].Value;
							var matchedWithConscription = terrritoryAddresses.Where(ap => ap.ConscriptionNumber == conscriptionNumber);
							if (matchedWithConscription.Count() == 1) {
								buildingAP = matchedWithConscription.Single();
							}
						}

						else if (IsBuildingWithProvisionalNumber(definitionPoint)) {
							string provisionalNumber = regexResult.Groups[0].Value;
							var matchedWithProvisional = terrritoryAddresses.Where(ap => ap.ProvisionalNumber == provisionalNumber);
							if (matchedWithProvisional.Count() == 1) {
								buildingAP = matchedWithProvisional.Single();
							}
						}

						if (buildingAP == null)
							continue;

						// Process matched AP
						buildingAP.Coordinates = definitionPoint.Coordinates;
						matchedAddressPoints.Add(buildingAP);
						terrritoryAddresses.Remove(buildingAP);
						territoryBuildings.Remove(definitionPoint);
					}

					foreach (var mappingDistrict in mapping.Districts) {
						unassignedAddressPoints[mappingDistrict].RemoveAll(toDelete => terrritoryAddresses.Contains(toDelete) == false);
					}
									
				} catch (System.Collections.Generic.KeyNotFoundException){
					Console.WriteLine("not found0");
					continue;		
				}

				}

			unmatchedAdressPoints = unassignedAddressPoints.SelectMany(ap => ap.Value).ToList();

			FilterUnmatchedDefinitionPoints(unassignedBuildngDefinitionPoints);
		}

		void FilterUnmatchedDefinitionPoints(Dictionary<string, List<BuildingDefinitionPoint>> unmatched) {
			foreach (var definitionPointsList in unmatched.Values) {
				foreach (var point in definitionPointsList.ToList()) {
					if (matchedAddressPoints.Any(ap => AreClose(ap.Coordinates, point.Coordinates))) {
						definitionPointsList.Remove(point);
					}
				}
			}

			foreach (var mapping in mappings.MappingsList) {
				try{
				if (mapping.Districts.Count > 1) {
					if (unmatchedBuildingDefinitionPoints.ContainsKey(District.emptyDistrict) == false) {
						unmatchedBuildingDefinitionPoints.Add(District.emptyDistrict, unmatched[mapping.TerritoryName]);
					}
					else {
						unmatchedBuildingDefinitionPoints[District.emptyDistrict].AddRange(unmatched[mapping.TerritoryName]);
					}

					continue;
				}


				if (unmatchedBuildingDefinitionPoints.ContainsKey(mapping.Districts[0]) == false) {
					unmatchedBuildingDefinitionPoints.Add(mapping.Districts[0], unmatched[mapping.TerritoryName]);
				}
				else {
					unmatchedBuildingDefinitionPoints[mapping.Districts[0]].AddRange(unmatched[mapping.TerritoryName]);
				}
				} catch (System.Collections.Generic.KeyNotFoundException){
					Console.WriteLine("not found3");
					continue;		
				}
			}
		}

		bool AreClose(PointGeo p1, PointGeo p2) {
			return Math.Abs(p1.Lat - p2.Lat) < 0.000002 && Math.Abs(p1.Lon - p2.Lon) < 0.000003;
		}

		bool IsBuildingWithConscriptionNumber(BuildingDefinitionPoint p) {
			string description = p.Description.Trim();
			if (description.StartsWith("č.p.") || description.StartsWith("čp.") || description.StartsWith("č.p")) {
				return true;
			}

			return false;
		}

		bool IsBuildingWithProvisionalNumber(BuildingDefinitionPoint p) {
			string description = p.Description.Trim();
			if (description.StartsWith("č.e.") || description.StartsWith("če.") || description.StartsWith("č.e")) {
				return true;
			}

			return false;
		}

		void Initialize(Dictionary<string, List<BuildingDefinitionPoint>> unassignedBuildngDefinitionPoints, Dictionary<District, List<AddressPoint>> unassignedAddressPoints) {
			foreach (var mapping in mappings.MappingsList) {
				if (unassignedBuildngDefinitionPoints.ContainsKey(mapping.TerritoryName) == false) {
					Territory territory;
					try{
						territory = territories.Territories.Where(t => t.Name.ToUpper(czechCulture) == mapping.TerritoryName.ToUpper(czechCulture)).Single();
					} catch (System.InvalidOperationException){
						Console.WriteLine("Mapping \"" + mapping.TerritoryName + "\" not found in territories");
						continue;		
					}
					
					List<BuildingDefinitionPoint> territoryPoints = points.GetPointsByTerritory(territory);

					foreach (var toRemove in territoryPoints.Where(buildng => buildng.Description.StartsWith("bez") ||
																																		buildng.Description.StartsWith("be2") ||
																																		buildng.Description.StartsWith("be5")).ToList()) {
						territoryPoints.Remove(toRemove);
					}

					unassignedBuildngDefinitionPoints.Add(mapping.TerritoryName, territoryPoints);

					foreach (var mappingDistrict in mapping.Districts) {
						if (unassignedAddressPoints.ContainsKey(mappingDistrict) == false) {
							Town districtTown = addresses.Regions.Where(r => r.Name.ToUpper(czechCulture) == mappingDistrict.RegionName.ToUpper(czechCulture)).Single()
																															 .Towns.Where(t => t.Name.ToUpper(czechCulture) == mappingDistrict.TownName.ToUpper(czechCulture)).Single();
							ElementWithStreets mappingAddressElement;
							if (mappingDistrict.TownDistrictName != null) {
								mappingAddressElement = districtTown.TownDistricts.Where(td => td.Name.ToUpper(czechCulture) == mappingDistrict.TownDistrictName.ToUpper(czechCulture)).Single();
							}
							else {
								mappingAddressElement = districtTown;
							}

							List<AddressPoint> districtAddressPoints = new List<AddressPoint>();
							districtAddressPoints.AddRange(mappingAddressElement.AddressPoints);

							foreach (var street in mappingAddressElement.Streets) {
								districtAddressPoints.AddRange(street.AddressPoints);
							}

							unassignedAddressPoints.Add(mappingDistrict, districtAddressPoints);
						}
					}

				}
			}
		}
	}
}
