using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Xml.Linq;

using CUZK.Common;

namespace CUZK.MergeDBWithPoints.Territories {
	class TerritoriesDB {
		public List<Territory> Territories;

		public TerritoriesDB() {
			Territories = new List<Territory>();
		}

		static double Distance(PointGeo p1, PointGeo p2) {
			return Math.Sqrt((p1.Lat - p2.Lat) * (p1.Lat - p2.Lat) + (p1.Lon - p2.Lon) * (p1.Lon - p2.Lon));
		}
		
		public static TerritoriesDB FromOsm(string filename) {
			TerritoriesDB result = new TerritoriesDB();

			XElement osmXml = XDocument.Load(filename).Root;

			var territoriesXml = osmXml.Elements("relation").Where(rel => rel.Elements("tag").Count() > 0 &&
																												     rel.Elements("tag").Any(tag => tag.Attribute("k").Value == "name")
																											 );
			Dictionary<string,XElement> wayDict = osmXml.Elements("way").ToDictionary(way => way.Attribute("id").Value,item => item);
			Dictionary<string,XElement> nodeDict = osmXml.Elements("node").ToDictionary(node => node.Attribute("id").Value,item => item);

			foreach (var territoryXml in territoriesXml) {
				Territory parsed = new Territory();
				parsed.Name = territoryXml.Elements("tag").Where(tag => tag.Attribute("k").Value == "name").Single().Attribute("v").Value;

				foreach (var wayXml in territoryXml.Elements("member").Where(way => way.Attribute("type").Value == "way")) {
					List<PointGeo> wayPoints = new List<PointGeo>();

					string wayID = wayXml.Attribute("ref").Value;
					foreach (var nodeRef in wayDict[wayID].Elements("nd")){
						string nodeId = nodeRef.Attribute("ref").Value;

						XElement nodeXml = nodeDict[nodeId];
						double lat = Double.Parse(nodeXml.Attribute("lat").Value, System.Globalization.CultureInfo.InvariantCulture.NumberFormat);
						double lon = Double.Parse(nodeXml.Attribute("lon").Value, System.Globalization.CultureInfo.InvariantCulture.NumberFormat);

						wayPoints.Add(new PointGeo(lat, lon));
					}

					if (parsed.Bounds.Count == 0) {
						parsed.Bounds.AddRange(wayPoints);
					}
					else {
						PointGeo lastPoint = parsed.Bounds.Last();
						if (Distance(lastPoint, wayPoints.First()) < Distance(lastPoint, wayPoints.Last())) {
							parsed.Bounds.AddRange(wayPoints);
						}
						else {
							wayPoints.Reverse();
							parsed.Bounds.AddRange(wayPoints);
						}
					}
				}
				

				parsed.Bounds.Remove(parsed.Bounds.Last());
				result.Territories.Add(parsed);
			}

			return result;

		}
	}
}
